@@ -26,6 +26,7 @@
 #include "util.h"
 #include "search.h"
 #include "status.h"
+#include "path.h"
 #include "gcc.h"
 
 #include <mpd/client.h>
@@ -56,6 +57,21 @@ static void my_finishCommand(struct mpd_connection *conn) {
 		printErrorAndExit(conn);
 }
 
+static void
+strip_trailing_slash(char *s)
+{
+	size_t len = strlen(s);
+
+	if (len == 0)
+		return;
+	--len;
+
+	if (s[len] == '/')
+		s[len] = '\0';
+
+	return;
+}
+
 SIMPLE_CMD(cmd_next, mpd_run_next, 1)
 SIMPLE_CMD(cmd_prev, mpd_run_previous, 1)
 SIMPLE_CMD(cmd_stop, mpd_run_stop, 1)
@@ -78,17 +94,37 @@ getStatus(struct mpd_connection *conn) {
 	return ret;
 }
 
+static bool
+contains_absolute_path(unsigned argc, char **argv)
+{
+	for (unsigned i = 0; i < argc; ++i)
+		if (argv[i][0] == '/')
+			return true;
+
+	return false;
+}
+
 int cmd_add (int argc, char ** argv, struct mpd_connection *conn )
 {
+	if (contains_absolute_path(argc, argv) && !path_prepare(conn))
+		printErrorAndExit(conn);
+
 	int i;
 
 	if (!mpd_command_list_begin(conn, false))
 		printErrorAndExit(conn);
 
 	for(i=0;i<argc;i++) {
+		strip_trailing_slash(argv[i]);
+
+		const char *path = argv[i];
+		const char *relative_path = to_relative_path(path);
+		if (relative_path != NULL)
+			path = relative_path;
+
 		if (options.verbosity >= V_VERBOSE)
-			printf("adding: %s\n", argv[i]);
-		mpd_send_add(conn, charset_to_utf8(argv[i]));
+			printf("adding: %s\n", path);
+		mpd_send_add(conn, charset_to_utf8(path));
 	}
 
 	if (!mpd_command_list_end(conn))
@@ -180,9 +216,7 @@ wait_current(struct mpd_connection *c)
 	if (mpd_connection_cmp_server_version(c, 0, 14, 0) < 0)
 		fprintf(stderr, "warning: MPD 0.14 required for this command\n");
 
-	struct mpd_status *status = mpd_run_status(c);
-	if (status == NULL)
-		printErrorAndExit(c);
+	struct mpd_status *status = getStatus(c);
 
 	const int old_song = get_active_song(status);
 	mpd_status_free(status);
@@ -193,10 +227,7 @@ wait_current(struct mpd_connection *c)
 		if (idle == 0)
 			printErrorAndExit(c);
 
-		status = mpd_run_status(c);
-		if (status == NULL)
-			printErrorAndExit(c);
-
+		status = getStatus(c);
 		new_song = get_active_song(status);
 		mpd_status_free(status);
 	} while (new_song == old_song);
@@ -319,6 +350,26 @@ int cmd_del ( int argc, char ** argv, struct mpd_connection *conn )
 	return 0;
 }
 
+int
+cmd_cdprev(mpd_unused int argc, mpd_unused char **argv,
+	   struct mpd_connection *conn)
+{
+	struct mpd_status *status;
+	status = getStatus(conn);
+
+	/* go to previous track if mpd is playing first 3 seconds of
+	   current track otherwise seek to beginning of current
+	   track */
+	if (mpd_status_get_elapsed_time(status) < 3) {
+		cmd_prev(0, NULL, conn);
+	} else {
+		if (!mpd_run_seek_id(conn, mpd_status_get_song_id(status), 0))
+			printErrorAndExit(conn);
+	}
+
+	return 1;
+}
+
 int
 cmd_toggle(mpd_unused int argc, mpd_unused char **argv, struct mpd_connection *conn)
 {
@@ -357,30 +408,126 @@ cmd_outputs(mpd_unused int argc, mpd_unused char **argv, struct mpd_connection *
 	return( 0 );
 }
 
-int
-cmd_enable(mpd_unused int argc, char **argv, struct mpd_connection *conn)
+static unsigned
+match_outputs(struct mpd_connection *conn,
+	      char **names, char **names_end, unsigned **ids_end)
 {
-	int arg;
+	struct mpd_output *output;
+	unsigned max = 0, *id = *ids_end;
 
-        if( ! parse_int( argv[0], &arg ) || arg <= 0 ) {
-		DIE( "Not a positive integer\n" );
+	mpd_send_outputs(conn);
+
+	while ((output = mpd_recv_output(conn)) != NULL) {
+		const char *name = mpd_output_get_name(output);
+		max = mpd_output_get_id(output);
+
+		for (char **n = names; n != names_end; ++n) {
+			if (!strcmp(*n, name)) {
+				*id = max;
+				++id;
+				*n = *names;
+				++names;
+				break;
+			}
+		}
+
+		mpd_output_free(output);
 	}
 
-	mpd_run_enable_output(conn, arg - 1);
-	return cmd_outputs(0, NULL, conn);
+	mpd_response_finish(conn);
+
+	for (char **n = names; n != names_end; ++n) {
+		fprintf(stderr, "%s: no such output\n", *n);
+	}
+
+	*ids_end = id;
+	return max;
+}
+
+static int
+enable_disable(int argc, char **argv, struct mpd_connection *conn,
+	       bool (*matched)(struct mpd_connection *conn, unsigned id),
+	       bool (*not_matched)(struct mpd_connection *conn, unsigned id))
+{
+	char **names = argv, **names_end = argv;
+	unsigned *ids, *ids_end, max;
+	bool only = false;
+	int arg;
+
+	if (!strcmp(argv[0], "only")) {
+		only = true;
+		++argv;
+		if (!--argc) {
+			DIE("No outputs specified.");
+		}
+	}
+
+	ids = malloc(argc * sizeof *ids);
+	ids_end = ids;
+
+	for (int i = argc; i; --i, ++argv) {
+		if (!parse_int(*argv, &arg)) {
+			*names_end = *argv;
+			++names_end;
+		} else if (arg <= 0) {
+			fprintf(stderr, "%s: not a positive integer\n", *argv);
+		} else {
+			/* We decrement by 1 to make it natural to the user. */
+			*ids_end++ = arg - 1;
+		}
+	}
+
+	if (only || names != names_end) {
+		max = match_outputs(conn, names, names_end, &ids_end);
+	}
+
+	if (ids == ids_end) {
+		goto done;
+	}
+
+	if (!mpd_command_list_begin(conn, false)) {
+		printErrorAndExit(conn);
+	}
+
+	if (only) {
+		for (unsigned i = 0; i <= max; ++i) {
+			bool found = false;
+			for (unsigned *id = ids;
+			     !found && id != ids_end;
+			     ++id) {
+				found = *id == i;
+			}
+			(found ? matched : not_matched)(conn, i);
+		}
+	} else {
+		for (unsigned *id = ids; id != ids_end; ++id) {
+			matched(conn, *id);
+		}
+	}
+
+	if (!mpd_command_list_end(conn) || !mpd_response_finish(conn)) {
+		printErrorAndExit(conn);
+	}
+
+	cmd_outputs(0, NULL, conn);
+
+done:
+	free(ids);
+	return 0;
+}
+
+int
+cmd_enable(int argc, char **argv, struct mpd_connection *conn)
+{
+	return enable_disable(argc, argv, conn, mpd_send_enable_output,
+			      mpd_send_disable_output);
 }
 
 int
 cmd_disable(mpd_unused int argc, char **argv, struct mpd_connection *conn)
 {
-	int arg;
-
-        if( ! parse_int( argv[0], &arg ) || arg <= 0 ) {
-		DIE( "Not a positive integer\n" );
-	}
-
-	mpd_run_disable_output(conn, arg - 1);
-	return cmd_outputs(0, NULL, conn);
+	return enable_disable(argc, argv, conn, mpd_send_disable_output,
+			      mpd_send_enable_output);
 }
 
 int cmd_play ( int argc, char ** argv, struct mpd_connection *conn )
@@ -597,20 +744,23 @@ int cmd_listall ( int argc, char ** argv, struct mpd_connection *conn )
 		listall = charset_to_utf8(argv[i]);
 
 	do {
+		char *tmp = strdup(listall);
+		strip_trailing_slash(tmp);
+
 		if (options.custom_format) {
-			if (!mpd_send_list_all_meta(conn, listall))
+			if (!mpd_send_list_all_meta(conn, tmp))
 				printErrorAndExit(conn);
 
 			print_entity_list(conn, MPD_ENTITY_TYPE_UNKNOWN);
 		} else {
-			if (!mpd_send_list_all(conn, listall))
+			if (!mpd_send_list_all(conn, tmp))
 				printErrorAndExit(conn);
 
 			print_filenames(conn);
 		}
 
 		my_finishCommand(conn);
-
+		free(tmp);
 	} while (++i < argc && (listall = charset_to_utf8(argv[i])) != NULL);
 
 	return 0;
@@ -618,6 +768,9 @@ int cmd_listall ( int argc, char ** argv, struct mpd_connection *conn )
 
 int cmd_update ( int argc, char ** argv, struct mpd_connection *conn)
 {
+	if (contains_absolute_path(argc, argv) && !path_prepare(conn))
+		printErrorAndExit(conn);
+
 	const char * update = "";
 	int i = 0;
 	unsigned id = 0;
@@ -628,7 +781,16 @@ int cmd_update ( int argc, char ** argv, struct mpd_connection *conn)
 	if(argc > 0) update = charset_to_utf8(argv[i]);
 
 	do {
-		mpd_send_update(conn, update);
+		char *tmp = strdup(update);
+		strip_trailing_slash(tmp);
+
+		const char *path = tmp;
+		const char *relative_path = to_relative_path(path);
+		if (relative_path != NULL)
+			path = relative_path;
+
+		mpd_send_update(conn, path);
+		free(tmp);
 	} while (++i < argc && (update = charset_to_utf8(argv[i])) != NULL);
 
 	if (!mpd_command_list_end(conn))
@@ -657,10 +819,7 @@ int cmd_update ( int argc, char ** argv, struct mpd_connection *conn)
 
 		/* determine the current "update id" */
 
-		status = mpd_run_status(conn);
-		if (status == NULL)
-			printErrorAndExit(conn);
-
+		status = getStatus(conn);
 		current_id = mpd_status_get_update_id(status);
 		mpd_status_free(status);
 
@@ -674,44 +833,38 @@ int cmd_update ( int argc, char ** argv, struct mpd_connection *conn)
 	return 1;
 }
 
-int cmd_ls ( int argc, char ** argv, struct mpd_connection *conn )
+static int
+ls_entity(int argc, char **argv, struct mpd_connection *conn,
+	  enum mpd_entity_type type)
 {
-	const char *ls;
+	const char *ls = "";
 	int i = 0;
 
 	if (argc > 0)
 		ls = charset_to_utf8(argv[i]);
-	else
-		ls = strdup("");
 
 	do {
 		if (!mpd_send_list_meta(conn, ls))
 			printErrorAndExit(conn);
 
-		print_entity_list(conn, MPD_ENTITY_TYPE_UNKNOWN);
+		print_entity_list(conn, type);
 		my_finishCommand(conn);
-
 	} while (++i < argc && (ls = charset_to_utf8(argv[i])) != NULL);
 
 	return 0;
 }
 
+int cmd_ls ( int argc, char ** argv, struct mpd_connection *conn )
+{
+	for (int i = 0; i < argc; i++)
+		strip_trailing_slash(argv[i]);
+
+	return ls_entity(argc, argv, conn, MPD_ENTITY_TYPE_UNKNOWN);
+}
+
 int cmd_lsplaylists ( int argc, char ** argv, struct mpd_connection *conn )
 {
-	const char * ls = "";
-	int i = 0;
-
-	if(argc>0) ls = charset_to_utf8(argv[i]);
-
-	do {
-		if (!mpd_send_list_meta(conn, ls))
-			printErrorAndExit(conn);
-
-		print_entity_list(conn, MPD_ENTITY_TYPE_PLAYLIST);
-		my_finishCommand(conn);
-
-	} while (++i < argc && (ls = charset_to_utf8(argv[i])) != NULL);
-	return 0;
+	return ls_entity(argc, argv, conn, MPD_ENTITY_TYPE_PLAYLIST);
 }
 
 int cmd_load ( int argc, char ** argv, struct mpd_connection *conn )
@@ -734,9 +887,7 @@ int cmd_load ( int argc, char ** argv, struct mpd_connection *conn )
 int cmd_insert (int argc, char ** argv, struct mpd_connection *conn )
 {
 	int ret;
-	struct mpd_status *status = mpd_run_status(conn);
-	if (status == NULL)
-		printErrorAndExit(conn);
+	struct mpd_status *status = getStatus(conn);
 
 	const int from = mpd_status_get_queue_length(status);
 
@@ -833,92 +984,52 @@ cmd_pause(mpd_unused int argc, mpd_unused char **argv, struct mpd_connection *co
 	return 1;
 }
 
-int cmd_repeat ( int argc, char ** argv, struct mpd_connection *conn )
+static int
+bool_cmd(int argc, char **argv, struct mpd_connection *conn,
+	 bool (*get_mode)(const struct mpd_status *status),
+	 bool (*run_set_mode)(struct mpd_connection *conn, bool mode))
 {
-	int mode;
+	bool mode;
 
-	if(argc==1) {
+	if (argc == 1) {
 		mode = get_boolean(argv[0]);
 		if (mode < 0)
 			return -1;
-	}
-	else {
+	} else {
 		struct mpd_status *status;
 		status = getStatus(conn);
-		mode = !mpd_status_get_repeat(status);
+		mode = !get_mode(status);
 		mpd_status_free(status);
 	}
 
-	if (!mpd_run_repeat(conn, mode))
+	if (!run_set_mode(conn, mode))
 		printErrorAndExit(conn);
 
 	return 1;
 }
 
-int cmd_random ( int argc, char ** argv, struct mpd_connection *conn )
+int cmd_repeat(int argc, char ** argv, struct mpd_connection *conn)
 {
-	int mode;
-
-	if(argc==1) {
-		mode = get_boolean(argv[0]);
-		if (mode < 0)
-			return -1;
-	}
-	else {
-		struct mpd_status *status;
-		status = getStatus(conn);
-		mode = !mpd_status_get_random(status);
-		mpd_status_free(status);
-	}
-
-	if (!mpd_run_random(conn, mode))
-		printErrorAndExit(conn);
-
-	return 1;
+	return bool_cmd(argc, argv, conn,
+			mpd_status_get_repeat, mpd_run_repeat);
 }
 
-int cmd_single ( int argc, char ** argv, struct mpd_connection *conn )
+int cmd_random(int argc, char ** argv, struct mpd_connection *conn)
 {
-	int mode;
-
-	if(argc==1) {
-		mode = get_boolean(argv[0]);
-		if (mode < 0)
-			return -1;
-	}
-	else {
-		struct mpd_status *status;
-		status = getStatus(conn);
-		mode = !mpd_status_get_single(status);
-		mpd_status_free(status);
-	}
-
-	if (!mpd_run_single(conn, mode))
-		printErrorAndExit(conn);
-
-	return 1;
+	return bool_cmd(argc, argv, conn,
+			mpd_status_get_random, mpd_run_random);
 }
 
-int cmd_consume ( int argc, char ** argv, struct mpd_connection *conn )
+int cmd_single(int argc, char ** argv, struct mpd_connection *conn)
 {
-	int mode;
+	return bool_cmd(argc, argv, conn,
+			mpd_status_get_single, mpd_run_single);
+}
 
-	if(argc==1) {
-		mode = get_boolean(argv[0]);
-		if (mode < 0)
-			return -1;
-	}
-	else {
-		struct mpd_status *status;
-		status = getStatus(conn);
-		mode = !mpd_status_get_consume(status);
-		mpd_status_free(status);
-	}
-
-	if (!mpd_run_consume(conn, mode))
-		printErrorAndExit(conn);
-
-	return 1;
+int cmd_consume(int argc, char ** argv, struct mpd_connection *conn)
+{
+	return bool_cmd(argc, argv, conn,
+			mpd_status_get_consume, mpd_run_consume);
 }
 
 int cmd_crossfade ( int argc, char ** argv, struct mpd_connection *conn )
@@ -1109,7 +1220,7 @@ static char * DHMS(unsigned long t)
 	t %= SECSPERMIN;
 	secs = t;
 
-	snprintf(buf, sizeof(buf) - 1, "%d days, %d:%02d:%02d",
+	snprintf(buf, sizeof(buf), "%d days, %d:%02d:%02d",
 	    days, hours, mins, secs);
 	return buf;
 }
